
{

  This is the code used for Winnow's Find Clusters feature.
  
  It is assumed that all particles in the system have the same mass, so it 
  won't work for a compound target.

  The dimers and trimers are identified on the basis of spatial proximity
  of 2 or 3 atoms, and (if ignore_PE is FALSE) their internal energy, Er.

  Er is the sum of COM kinetic energy and potential energy calculated via
  a Morse potential.

  The routine is not very sophisticated and relies on the fact that N-mers
  appear with order of magnitude differences in yield. So, for example, atoms
  registered as part of a trimer will also be registered as part of a trimer.
  Similar problems arise with multimers.

  A better algorithm may be found in S.D. Stoddard, J. Comput. Phys. 27 (1978)
  291.

  I recommend that users of Kalypso use their own implementation of Stoddard's
  algorithm, rather than this one, to count clusters.

}

unit clusters;

interface

uses SnkDefs;

procedure TestClusters(SnkName,Datname,Outfile: ShortString;D,alfa,r0,CutOff:single;
ignore_PE,Count_Multimers:boolean;Bulkatom:integer);

implementation

uses SysUtils,SKMsgs,Forms,Queries;

const
  MAXDATA = MAXSNK;  { max. no. of records in SNK file }
  ep = 1.602177E-19; { conversion factor for eV }

type
  TSingleArray = array[1..MAXDATA] of single;
  PSingleArray = ^ TSingleArray;
type
  TIntArray = array[1..MAXDATA] of integer;
  PIntArray = ^ TIntArray;

procedure TestClusters(SnkName,Datname,Outfile: ShortString;D,alfa,r0,CutOff:single;
ignore_PE,Count_Multimers:boolean;Bulkatom:integer);

function V(r: single):single; { pair potential }
begin
  if r > CUTOFF then V := 0.0
  else V := D*exp(-2.0*alfa*(r-r0)) - 2.0*D*exp(-alfa*(r-r0));
end;

var
  f,g: TSnkFile;
  h: text;
  Rc: TSnookRec;
  ndata,n,i,j,ndimer,ntrimer,k: integer;
  x,y,z,px,py,pz: PSingleArray;
  idx,run,row: PIntArray;
  r,pr2,Er,mass,U,Ebary,rij,rik,rjk,KE, KE_com: single;
  InfoStr: ShortString;
  _break: boolean;
begin
  D := D*ep;                { convert to SI units }
  alfa := alfa*1.0E10;
  r0 := r0*1.0E-10;
  CUTOFF := CUTOFF*1.0E-10; { potential truncation distance in metres }

  assign(f,SnkName);        { open files }
  reset(f);
  assign(g,DatName);
  rewrite(g);
  assign(h,Outfile);
  rewrite(h);
  ndata := filesize(f);                { number of recs in SNK file }
  GetMem(idx,ndata*sizeOf(integer));   { allocate mem }
  GetMem(run,ndata*sizeOf(integer));
  GetMem(row,ndata*sizeOf(integer));
  GetMem(x,ndata*sizeOf(single));
  GetMem(y,ndata*sizeOf(single));
  GetMem(z,ndata*sizeOf(single));
  GetMem(px,ndata*sizeOf(single));
  GetMem(py,ndata*sizeOf(single));
  GetMem(pz,ndata*sizeOf(single));

  { read data }
  n := 0;
  while not eof(f) do
    begin
      read(f,Rc);
      inc(n);
      x^[n] := Rc.rx;
      y^[n] := Rc.ry;
      z^[n] := Rc.rz;
      px^[n] := Rc.px;
      py^[n] := Rc.py;
      pz^[n] := Rc.pz;
      idx^[n] := Rc.rw;
      run^[n] := Rc.rn;
      row^[n] := Rc.rn;
    end;

  mass := Rc.ms;  { assumed the same for all particles }

  { start dimers counting proc }
  ndimer := 0;
  for i := 1 to ndata-1 do
    for j := i+1 to ndata do
      begin
        if run^[i] <> run^[j] then break;   { cluster atoms must appear in same run }

        r := sqrt(sqr(x^[i]-x^[j]) + sqr(y^[i]-y^[j]) + sqr(z^[i]-z^[j]));
        pr2 := sqr(px^[i]-px^[j]) + sqr(py^[i]-py^[j]) + sqr(pz^[i]-pz^[j]);
        U := V(r);
        Er := pr2/(4.0*mass) + U; { include reduced mass ! Only valid for atoms of same type }

        if (r < CUTOFF) and ( (Er < 0.0) or IGNORE_PE) then  { record hit }
          begin
            { flag dimers which include an atom from the bulk }
            if (idx^[i] <= BULKATOM) and (idx^[j] <= BULKATOM) then writeln(h,idx^[i],' ',idx^[j])
            else writeln(h,idx^[i],' ',idx^[j],' ***'); { flagged }
            inc(ndimer);
              begin   { report dimer info }
                {
                  write momenta in such a way that Winnow computes the
                  correct KE for the cluster COM:
                  Tcom = 0.5(p1+p2)**2/4m = 0.5(Rc.p)**2/2m

                  so Rc.p = (p1+p2)/sqrt(2)
                }
                Rc.px := (px^[i]+px^[j])/sqrt(2);
                Rc.py := (py^[i]+py^[j])/sqrt(2);
                Rc.pz := (pz^[i]+pz^[j])/sqrt(2);
                Rc.rx := 0.5*(x^[i]+x^[j]);
                Rc.ry := 0.5*(y^[i]+y^[j]);
                Rc.rz := 0.5*(z^[i]+z^[j]);
                write(g,Rc);
              end;
            { break when first interaction is detected, or not? }
            if not COUNT_MULTIMERS then break;
          end;
      end;

  InfoStr := 'Recs: '+IntToStr(ndata)+', Dimers: '+IntToStr(ndimer);
  Application.MainForm.Perform(wm_WriteMsg,0,longint(@InfoStr));
  { end dimers counting proc }

  { start trimers counting proc }

  _break := false;
  { count trimers }
  ntrimer := 0;
  for i := 1 to ndata-2 do
    for j := i+1 to ndata-1 do
    begin
      if _break then   // flag break in inner loop
        begin
          _break := false;
          break;
        end;
      if (run^[i] <> run^[j]) then break;
      rij := sqrt(sqr(x^[i]-x^[j]) + sqr(y^[i]-y^[j]) + sqr(z^[i]-z^[j]));
      if rij > CUTOFF then continue;
      for k := j+1 to ndata do
        begin
          if (run^[j] <> run^[k]) then break;
          rik := sqrt(sqr(x^[i]-x^[k]) + sqr(y^[i]-y^[k]) + sqr(z^[i]-z^[k]));
          if rik > CUTOFF then continue;
          rjk := sqrt(sqr(x^[k]-x^[j]) + sqr(y^[k]-y^[j]) + sqr(z^[k]-z^[j]));
          if rjk > CUTOFF then continue;

          { total KE in lab frame }
          KE := 0.5*(sqr(px^[i])+sqr(py^[i])+sqr(pz^[i])+sqr(px^[j])+sqr(py^[j])+
                     sqr(pz^[j])+sqr(px^[k])+sqr(py^[k])+sqr(pz^[k]))/(mass);

          { KE associated with barycentre = KE(total) - KE(in com) }
          KE_com := 0.5*(sqr(px^[i]+px^[j]+px^[k])+ sqr(py^[i]+py^[j]+py^[k])+ sqr(pz^[i]+pz^[j]+pz^[k]))/(3.0*mass);

          { total potential E }
          U := V(rij) + V(rik) + V(rjk);
          Er := KE-KE_com + U; 

          { test for trimer formation }
          if (Er < 0.0) or IGNORE_PE then
            begin
              inc(ntrimer);
              writeln(h,i,' ',j,' ',k,' ','trimer bond lengths: ',rij*1E10:8:4,rik*1E10:8:4,rjk*1E10:8:4);
              { put other output code here }
            end;
          if not COUNT_MULTIMERS then
            begin
              _break := true;
              break
            end;
        end;   { for k }
    end;       { for j, i }

  InfoStr := 'Recs: '+IntToStr(ndata)+', Trimers: '+IntToStr(ntrimer);
  Application.MainForm.Perform(wm_WriteMsg,0,longint(@InfoStr));  { output for screen }

  { end trimers counting proc }

  FreeMem(x,ndata*sizeOf(single));
  FreeMem(y,ndata*sizeOf(single));
  FreeMem(z,ndata*sizeOf(single));
  FreeMem(px,ndata*sizeOf(single));
  FreeMem(py,ndata*sizeOf(single));
  FreeMem(pz,ndata*sizeOf(single));
  FreeMem(idx,ndata*sizeOf(integer));
  FreeMem(run,ndata*sizeOf(integer));
  FreeMem(row,ndata*sizeOf(integer));
  close(f);
  close(g);
  close(h);
end;

end.