program ScatterCount;

{ Compiled with Turbo Pascal (7.0) }
{
  This is an example of a quick-and-dirty solution to a SNK file
  processing problem:

  Command line syntax:

  scat-cnt num

  where num (optional parameter) is a number (0.5 by default)

  The program will search for all SNK files named 'XX.SNK',
  where XX is some number (e.g.45.SNK). The program counts the number
  of records in the file where the particle is travelling within
  0.5 degrees of the altitudinal angle given by the stem XX
  (e.g. 450.5 degrees).

  It also sorts the records out according to whether they correspond to
  motion in the +x or -x directions respectively.

  Assumption: the SNK file contains only records from ejected
  projectile particles - assured by the corresponding output
  option in SPIDER.

  Use: if no command-line argument is given, the program will
  use a 0.5 degree counting interval. However, a new interval can
  be invoked on the command line: "scat-cnt 1.5" (for 1.5 degree
  interval).

  This is a typical problem that arises in processing ICISS
  simulations. You can test this program out on the output
  files generated by the \EXAMPLES\ICISS project. You can set
  DELTA (the angular interval) to a value other than 0.5 if
  you wish. The program will process ALL files with valid names
  (i.e. XX.SNK, where XX is a numeric character string) in the
  working directory.

  Obviously this trick only works if the output SNK file name
  corresponds to the ejection angle in which you are interested.
  (But don't forget that you can always rename the SNK file.)

  You can achieve the same result using Winnow's filter option,
  but more care and time are needed.

  If you write your own processing routines, you will most likely
  find that most of the code used here can be retained, except
  for that contained in the block:

       with R do
         begin
           [Your unique code will go here]
         end;

}


uses Dos;
label
  NEXT;
const
  DELTA:single = 0.5;     { angular binning range, in degrees }
type
  PSnookRec = ^TSnookRec;
  TSnookRec = record
    rx,ry,rz,px,py,pz,ti,ms:single;
    rw,rn: longint;
  end;
type
  TSnkFile = file of TSnookRec;
var
 DirInfo: SearchRec;
 S: PathStr;
 R: TSnookRec;
 alt,bin: single;
 g: text;
 f: TSnkFile;
 angle,code,pos_angle,neg_angle: integer;
 cmdstr: string[10];
begin
 if ParamCount = 0 then
   writeln('Counting recs in 0.5 bins')
 else if ParamCount = 1 then
   begin
     val(ParamStr(1),bin,code);
     if code <> 0 then
       writeln('Invalid command line (must be number): Counting recs in 0.5 bins')
     else
       begin
         DELTA := bin;
         writeln('Counting recs in ',DELTA:5:3,' bins')
       end
   end
 else writeln('Invalid command line (must be one number): Counting recs in 0.5 bins');

 assign(g,'angle$.dat');       { this is the name of the output file }
 rewrite(g);
 FindFirst('*.SNK', Archive, DirInfo);
 while DosError = 0 do
 begin
   pos_angle := 0;
   neg_angle := 0;
   write(DirInfo.Name);
   S := DirInfo.Name;
   dec(S[0],4);
   val(S,angle,code);            { this is the angle to search for }
   if code <> 0 then
     begin
       writeln(' (not processed: non-numeric name)');
       goto NEXT;  { non-numeric file name }
     end
   else writeln;
   assign(f,DirInfo.Name);
   reset(f);
   write(g,S,' ');
   while not eof(f) do
     begin
       read(f,R);                { read each record, and test it }
       with R do
         begin
           if (px > 0.0) or (py > 0.0) then { avoid divide by zero }
             alt := 180/pi*arctan(pz/sqrt(px*px+py*py))
           else alt := 90.0;
           if (alt < angle + DELTA) and (alt > angle - DELTA) then
             begin
               if px > 0.0 then inc(pos_angle)
               else inc(neg_angle);
             end;
         end;
     end;
   close(f);
   writeln(g,pos_angle,' ',neg_angle);
NEXT:
   FindNext(DirInfo);
 end;
 close(g);
 writeln;
 writeln('Output written to file: "angle$.dat"');
end.